import asyncio

from .constants import *
from .wireformat import *
from .hosts import HostsResolver
from .system import SystemResolver
from .mdns import MulticastResolver

class SmartResolver(object):

    def __init__(self):
        self.sys = SystemResolver()
        self.hosts = HostsResolver()
        self.mdns = MulticastResolver()

    def lookup(self, query, prefer_ipv6=None, should_cache=True, recursive=False):
        outer = asyncio.Future()

        # First, try looking in /etc/hosts
        f = self.hosts.lookup(query)

        def callback(f):
            if f.cancelled():
                outer.cancel()
                return

            exc = f.exception()
            if exc is not None:
                outer.set_exception(exc)
                return

            reply = f.result()

            if reply.rcode == NXDOMAIN:
                if query.name.endswith(b'.local'):
                    f2 = self.mdns.lookup(query, use_ipv6=prefer_ipv6)
                else:
                    f2 = self.sys.lookup(query, prefer_ipv6,
                                         should_cache, recursive)

                def cb2(f):
                    if f.cancelled():
                        outer.cancel()
                        return

                    exc = f.exception()
                    if exc is not None:
                        outer.set_exception(exc)
                        return

                    outer.set_result(f.result())

                f2.add_done_callback(cb2)
            else:
                outer.set_result(reply)

        f.add_done_callback(callback)

        return outer



# These are for DNS messages
QR          = 0x8000
OPCODE_MASK = 0x7800
QUERY       = 0x0000
IQUERY      = 0x0800
STATUS      = 0x1000
#           = 0x1800
NOTIFY      = 0x2000
UPDATE      = 0x2800

AA          = 0x0400
TC          = 0x0200
RD          = 0x0100
RA          = 0x0080
Z           = 0x0040
AD          = 0x0020
CD          = 0x0010

DO          = 0x8000

RCODE_MASK  = 0x000f



#
# A Darwin-specific SystemResolver implementation that watches the system
# configuration store to spot changes to the DNS server settings.
#
# Uses ctypes, rather than using C code, so this is still pure Python :-)
#
import ipaddress
from ctypes import *

from .resolver import Resolver, RoundRobinServer, RandomServer

sc = cdll.LoadLibrary('/System/Library/Frameworks/SystemConfiguration.framework/SystemConfiguration')
cf = cdll.LoadLibrary('/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation')
libSystem = cdll.LoadLibrary('/usr/lib/libSystem.dylib')

class _CFString(Structure):
    pass
class _CFDictionary(Structure):
    pass
class _CFArray(Structure):
    pass
class _CFAllocator(Structure):
    pass
class _CFType(Structure):
    pass

CFAllocatorRef = POINTER(_CFAllocator)
CFDictionaryRef = POINTER(_CFDictionary)
CFStringRef = POINTER(_CFString)
CFArrayRef = POINTER(_CFArray)
CFTypeRef = c_void_p
CFIndex = c_uint64

CFArrayRetainCallBack = POINTER(CFUNCTYPE(c_void_p, CFAllocatorRef, c_void_p))
CFArrayReleaseCallBack = POINTER(CFUNCTYPE(None, CFAllocatorRef, c_void_p))
CFArrayCopyDescriptionCallBack = POINTER(CFUNCTYPE(CFStringRef, c_void_p))
CFArrayEqualCallBack = POINTER(CFUNCTYPE(c_bool, c_void_p, c_void_p))

class CFArrayCallBacks(Structure):
    _fields_ = [("version", CFIndex),
                ("retain", CFArrayRetainCallBack),
                ("release", CFArrayReleaseCallBack),
                ("copyDescription", CFArrayCopyDescriptionCallBack),
                ("equal", CFArrayEqualCallBack)]

kCFTypeArrayCallBacks = CFArrayCallBacks.in_dll(cf, 'kCFTypeArrayCallBacks')

kCFStringEncodingUTF8 = 0x08000100

CFRetain = cf.CFRetain
CFRelease = cf.CFRelease
CFShow = cf.CFShow
CFRetain.argtypes = [c_void_p]
CFRelease.argtypes = [c_void_p]
CFShow.argtypes = [c_void_p]

CFStringCreateWithCString = cf.CFStringCreateWithCString
CFStringCreateWithCString.restype = CFStringRef
CFStringCreateWithCString.argtypes = [c_void_p, c_char_p, c_uint]

CFStringGetCStringPtr = cf.CFStringGetCStringPtr
CFStringGetCStringPtr.restype = c_char_p
CFStringGetCStringPtr.argtypes = [CFStringRef, c_uint]

CFStringGetCString = cf.CFStringGetCString
CFStringGetCString.restype = c_bool
CFStringGetCString.argtypes = [CFStringRef, c_char_p, c_size_t, c_uint]

CFArrayCreate = cf.CFArrayCreate
CFArrayCreate.restype = CFArrayRef
CFArrayCreate.argtypes = [CFAllocatorRef, POINTER(c_void_p), CFIndex,
                             POINTER(CFArrayCallBacks)]

CFArrayGetCount = cf.CFArrayGetCount
CFArrayGetCount.restype = CFIndex
CFArrayGetCount.argtypes = [CFArrayRef]

CFArrayGetValueAtIndex = cf.CFArrayGetValueAtIndex
CFArrayGetValueAtIndex.restype = CFTypeRef
CFArrayGetValueAtIndex.argtypes = [CFArrayRef, CFIndex]

CFDictionaryGetValue = cf.CFDictionaryGetValue
CFDictionaryGetValue.restype = CFTypeRef
CFDictionaryGetValue.argtypes = [CFDictionaryRef, CFTypeRef]

copyDescription_t = CFUNCTYPE(CFStringRef, c_void_p)
retain_t = CFUNCTYPE(c_void_p, c_void_p)
release_t = CFUNCTYPE(None, c_void_p)

class _SCDynamicStore(Structure):
    pass
SCDynamicStoreRef = POINTER(_SCDynamicStore)

class SCDynamicStoreContext(Structure):
    _fields_ = [("version", CFIndex),
                ("info", py_object),
                ("retain", POINTER(retain_t)),
                ("release", POINTER(release_t)),
                ("copyDescription", POINTER(copyDescription_t))]

SCDynamicStoreCallBack = CFUNCTYPE(None, SCDynamicStoreRef, CFArrayRef, c_void_p)

SCDynamicStoreCreate = sc.SCDynamicStoreCreate
SCDynamicStoreCreate.restype = SCDynamicStoreRef
SCDynamicStoreCreate.argtypes = [CFAllocatorRef, CFStringRef, SCDynamicStoreCallBack, POINTER(SCDynamicStoreContext)]

class _dispatch_queue(Structure):
    pass

dispatch_queue_t = POINTER(_dispatch_queue)

dispatch_get_global_queue = libSystem.dispatch_get_global_queue
dispatch_get_global_queue.restype = dispatch_queue_t
dispatch_get_global_queue.argtypes = [c_long, c_ulong]

DISPATCH_QUEUE_PRIORITY_LOW = -2

SCDynamicStoreSetDispatchQueue = sc.SCDynamicStoreSetDispatchQueue
SCDynamicStoreSetDispatchQueue.restype = c_bool
SCDynamicStoreSetDispatchQueue.argtypes = [SCDynamicStoreRef, dispatch_queue_t]

SCDynamicStoreSetNotificationKeys = sc.SCDynamicStoreSetNotificationKeys
SCDynamicStoreSetNotificationKeys.restype = c_bool
SCDynamicStoreSetNotificationKeys.argtypes = [SCDynamicStoreRef, CFArrayRef,
                                              CFArrayRef]

SCDynamicStoreCopyValue = sc.SCDynamicStoreCopyValue
SCDynamicStoreCopyValue.restype = CFTypeRef
SCDynamicStoreCopyValue.argtypes = [SCDynamicStoreRef, CFStringRef]

def CFSTR(s):
    return CFStringCreateWithCString(None, s.encode('utf8'),
                                     kCFStringEncodingUTF8)

_python_asyncdns = CFSTR('python-asyncdns')
_state_network_global_dns = CFSTR('State:/Network/Global/DNS')
_server_addresses = CFSTR('ServerAddresses')

def CFStringGetPythonString(cfstr):
    s = CFStringGetCStringPtr(cfstr, kCFStringEncodingUTF8)
    if s is None:
        tmp_size = 512
        tmp = create_string_buffer(tmp_size)
        if CFStringGetCString(cfstr, tmp, tmp_size, kCFStringEncodingUTF8):
            s = tmp.value
    return s

def _callback(store, changedKeys, context):
    resolver = cast(context, py_object).value
    resolver.servers_changed()

class SystemResolver(Resolver):

    def __init__(self):
        self._servers = None
        ctx = SCDynamicStoreContext(0, self, None, None, None)
        self._callback = SCDynamicStoreCallBack(_callback)
        self._dynamic_store = SCDynamicStoreCreate(None, _python_asyncdns,
                                                   self._callback,
                                                   byref(ctx))
        queue = dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_LOW, 0)
        SCDynamicStoreSetDispatchQueue(self._dynamic_store, queue)
        super(SystemResolver, self).__init__()
        self._start_watching()
        self._update_servers()

    def _start_watching(self):
        keys = (CFStringRef * 1)()
        keys[0] = _state_network_global_dns
        array = CFArrayCreate(None, cast(keys, POINTER(c_void_p)),
                              1, byref(kCFTypeArrayCallBacks))
        CFRelease(keys[0])
        SCDynamicStoreSetNotificationKeys(self._dynamic_store, array, None)
        CFRelease(array)

    def _stop_watching(self):
        SCDynamicStoreSetNotificationKeys(self._dynamic_store, None, None)

    def _update_servers(self):
        server_list = []
        dns = cast(SCDynamicStoreCopyValue(self._dynamic_store,
                                           _state_network_global_dns),
                   CFDictionaryRef)
        servers = cast(CFDictionaryGetValue(dns, _server_addresses),
                       CFArrayRef)
        count = CFArrayGetCount(servers)

        for n in range(0, count):
            server = cast(CFArrayGetValueAtIndex(servers, n), CFStringRef)
            s = CFStringGetPythonString(server)

            if s is not None:
                try:
                    s_ip = ipaddress.ip_address(s.decode('utf8'))
                    server_list.append((s_ip, 53))
                except ValueError:
                    pass

        CFRelease(dns)

        self._servers = RoundRobinServer(server_list)

    def __del__(self):
        CFRelease(self._dynamic_store)

    def servers_changed(self):
        self._update_servers()

    def lookup(self, query, prefer_ipv6=False,
               should_cache=True, recursive=False):
        return super(SystemResolver, self).lookup(query, self._servers,
                                                  prefer_ipv6, should_cache,
                                                  recursive)



import struct
import datetime

from . import constants, utils
from .rr import rr, RR

@rr(constants.DNSKEY, constants.ANY)
class DNSKEY(RR):
    def __init__(self, name, rr_class, ttl, flags, protocol, algorithm, key):
        super(DNSKEY, self).__init__(name, constants.DNSKEY, rr_class, ttl)
        self.flags = flags
        self.protocol = protocol
        self.algorithm = algorithm
        self.key = key

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}\t{}\t{}\t({})'\
            .format(utils.escape_string(self.name),
                    self.ttl,
                    utils.rrclass_to_string(self.rr_class),
                    utils.rrtype_to_string(self.rr_type),
                    self.flags,
                    self.protocol,
                    self.algorithm,
                    utils.base64(self.key))

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        flags, protocol, algorithm = struct.unpack(b'>HBB', packet[ptr:ptr+4])
        key = packet[ptr+4:ptr+rdlen]
        return DNSKEY(name, rr_class, ttl, flags, protocol, algorithm, key)

@rr(constants.RRSIG, constants.ANY)
class RRSIG(RR):
    def __init__(self, name, rr_class, ttl, type_covered, algorithm, labels,
                 orig_ttl, sig_expiration, sig_inception, key_tag, signer_name,
                 signature):
        super(RRSIG, self).__init__(name, constants.RRSIG, rr_class, ttl)
        self.type_covered = type_covered
        self.algorithm = algorithm
        self.labels = labels
        self.orig_ttl = orig_ttl
        self.sig_expiration = sig_expiration
        self.sig_inception = sig_inception
        self.key_tag = key_tag
        self.signer_name = signer_name
        self.signature = signature

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t({})'\
            .format(utils.escape_string(self.name), self.ttl,
                    utils.rrclass_to_string(self.rr_class),
                    utils.rrtype_to_string(self.rr_type),
                    self.type_covered,
                    self.algorithm,
                    self.labels,
                    self.orig_ttl,
                    self.sig_expiration.strftime('%Y%m%d%H%M%S'),
                    self.sig_inception.strftime('%Y%m%d%H%M%S'),
                    self.key_tag,
                    utils.escape_string(self.signer_name),
                    utils.base64(self.signature))

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        tc, alg, labels, orig_ttl, sig_exp, sig_inc, key_tag \
            = struct.unpack(b'>HBBLLLH', packet[ptr:ptr+18])
        signer, pt2 = utils.decode_domain(packet, ptr+18)
        signature = packet[pt2:ptr+rdlen]
        sig_exp = datetime.datetime.fromtimestamp(sig_exp)
        sig_inc = datetime.datetime.fromtimestamp(sig_inc)
        return RRSIG(name, rr_class, ttl, tc, alg, labels, orig_ttl,
                     sig_exp, sig_inc, key_tag, signer, signature)

@rr(constants.NSEC, constants.ANY)
class NSEC(RR):
    def __init__(self, name, rr_class, ttl, next_domain, types):
        super(NSEC, self).__init__(name, constants.NSEC, rr_class, ttl)
        self.next_domain = next_domain
        self.types = types

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}\t({})'\
            .format(utils.escape_string(self.name), self.ttl,
                    utils.rrclass_to_string(self.rr_class),
                    utils.rrtype_to_string(self.rr_type),
                    utils.escape_string(self.next_domain),
                    ', '.join([utils.rrtype_to_string(t) for t in self.types]))

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        end = ptr + rdlen
        next_domain, ptr = utils.decode_domain(packet, ptr)
        types = set()
        while ptr < end:
            window = packet[ptr]
            wlen = packet[ptr + 1]
            ptr += 2

            bmp = packet[ptr:ptr+wlen]

            for n,b in enumerate(bmp):
                mask = 0x1
                for m in range(0, 8):
                    if b & mask:
                        t = n + m
                        if t not in (constants.OPT, constants.IXFR,
                                     constants.AXFR, constants.ANY):
                            types.add(n + m)
                    mask <<= 1
        return NSEC(name, rr_class, ttl, next_domain, types)

@rr(constants.DS, constants.ANY)
class DS(RR):
    def __init__(self, name, rr_class, ttl, key_tag, algorithm,
                 digest_type, digest):
        super(DS, self).__init__(name, constants.DS, rr_class, ttl)
        self.key_tag = key_tag
        self.algorithm = algorithm
        self.digest_type = digest_type
        self.digest = digest

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}\t{}\t{}\t({})'\
            .format(utils.escape_string(self.name), self.ttl,
                    utils.rrclass_to_string(self.rr_class),
                    utils.rrtype_to_string(self.rr_type),
                    self.key_tag,
                    self.algorithm,
                    self.digest_type,
                    utils.base64(self.digest))

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        key_tag, algorithm, digest_type \
            = struct.unpack(b'>HBB', packet[ptr:ptr+4])
        digest = packet[4:]
        return DS(name, rr_class, ttl, key_tag, algorithm, digest_type, digest)



import asyncio

class Timeout(object):
    def __init__(self, timeout, callback):
        loop = asyncio.get_event_loop()
        loop.call_later(timeout, self.fire)
        self.callback = callback
        self.cancelled = False

    def fire(self):
        if self.cancelled:
            return
        self.callback()

    def cancel(self):
        self.cancelled = True
        self.callback = None




#
# Python-based async DNS implementation
#

import asyncio
import struct
import time
import random
import weakref
import socket
import ipaddress
import encodings.idna
import sys
import pkg_resources
import re
import errno

from . import rr
from .constants import *
from .utils import decode_domain, decode_pascal_string, rcode_to_string, \
    build_dns_packet, MAX_PACKET_SIZE
from .wireformat import *
from .timeout import Timeout

TIMEOUT         = 30
MAX_TRIES       = 5
MAX_TTL         = 3 * 3600

_rng = random.SystemRandom()

# Load and parse the named.root file
_ipv4_roots = []
_ipv6_roots = []

_ipv4_re = re.compile('[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}')

_named_root = pkg_resources.resource_string('asyncdns', 'named.root')\
                           .decode('ascii')
_space_re = re.compile(r'\s+')
for line in _named_root.splitlines():
    line = line.strip()
    if line.startswith(';'):
        continue

    line = _space_re.split(line)
    if line[2] == 'A':
        addr = ipaddress.ip_address(line[3])
        _ipv4_roots.append((addr, 53))
    elif line[2] == 'AAAA':
        addr = ipaddress.ip_address(line[3])
        _ipv6_roots.append((addr, 53))

_all_roots = _ipv4_roots + _ipv6_roots

class Query(object):
    """Represents a DNS query."""

    def __init__(self, name, q_type, q_class):
        # The name being queried
        if isinstance(name, ipaddress.IPv4Address):
            name = b'.'.join(str(name).encode('ascii').split(b'.')[::-1]) \
                   + b'.in-addr.arpa'
            self.name = name
        elif isinstance(name, ipaddress.IPv6Address):
            name = '.'.join(name.exploded.replace(':','')[::-1]).encode('ascii')\
                   + b'.ip6.arpa'
            self.name = name
        elif isinstance(name, str):
            name = name.rstrip('.')
            if name == '':
                self.name = b''
            else:
                self.name = b'.'.join([encodings.idna.ToASCII(label)
                                       for label in name.split('.')])
        else:
            name = name.rstrip(b'.')
            self.name = name

        # The query type (e.g. A)
        self.q_type = q_type

        # The query class (e.g. IN)
        self.q_class = q_class

    def __lt__(self, other):
        return (self.name < other.name
                or (self.name == other.name
                    and (self.q_type < other.q_type
                         or (self.q_type == other.q_type
                             and self.q_class < other.q_class))))

    def __eq__(self, other):
        return (self.name == other.name
                and self.q_type == other.q_type
                and self.q_class == other.q_class)

    def __ne__(self, other):
        return (self.name != other.name
                or self.q_type != other.q_type
                or self.q_class != other.q_class)

    def __gt__(self, other):
        return other < self

    def __ge__(self, other):
        return not self < other

    def __le__(self, other):
        return not other < self

    def __hash__(self):
        return hash((self.name, self.q_type, self.q_class))

    def __repr__(self):
        return 'Query({!r}, {!r}, {!r})'.format(self.name,
                                                self.q_type,
                                                self.q_class)

class Reply(object):
    def __init__(self, flags, rcode, answers, authorities, additional):
        self.flags = flags
        self.rcode = rcode
        self.answers = answers
        self.authorities = authorities
        self.additional = additional

    def update_ttls(ttl):
        for answer in self.answers:
            answer.ttl = ttl
        for auth in self.authorities:
            auth.ttl = ttl
        for add in self.additional:
            add.ttl = ttl

    def __str__(self):
        flags = []
        for name,mask in (('AA', AA), ('TC', TC), ('RD', RD), ('RA', RA),
                          ('Z', Z), ('AD', AD), ('CD', CD)):
            if self.flags & mask:
                flags.append(name)

        header = ';; {} ({})'.format(rcode_to_string(self.rcode), ', '.join(flags))

        lines = [header]
        lines.append('; {} answers:'.format(len(self.answers)))
        for answer in self.answers:
            lines.append(str(answer))
        lines.append('; {} authorities:'.format(len(self.authorities)))
        for auth in self.authorities:
            lines.append(str(auth))
        lines.append('; {} additional:'.format(len(self.additional)))
        for add in self.additional:
            lines.append(str(add))
        return '\n'.join(lines)

class BadReply(Exception):
    pass

class RoundRobinServer(object):
    def __init__(self, servers):
        self.servers = servers
        self.server_ndx = 0

    def __next__(self):
        server = self.servers[self.server_ndx]
        self.server_ndx = (self.server_ndx + 1) % len(self.servers)
        return server

class RandomServer(object):
    def __init__(self, servers):
        self.servers = servers

    def __next__(self):
        server_ndx = random.randrange(0, len(self.servers))
        return self.servers[server_ndx]

class DNSProtocol(object):

    def __init__(self, resolver, query, server_selector,
                 uid, using_tcp, prefer_ipv6, should_cache, recursive):
        self.resolver = weakref.ref(resolver)
        self.query = query
        self.server_selector = server_selector
        self.server = None
        self.uid = uid
        self.using_tcp = using_tcp
        self.prefer_ipv6 = prefer_ipv6
        self.should_cache = should_cache
        self.recursive = recursive
        self.transport = None
        self._retry_count = 0
        self._waiters = []
        self._timeout = None

    def set_timeout(self, timeout):
        self.cancel_timeout()
        self._timeout = Timeout(timeout, self.timed_out)

    def cancel_timeout(self):
        if self._timeout is not None:
            self._timeout.cancel()
            self._timeout = None

    def bind_random_port(self, sock):
        while True:
            port = _rng.randrange(1024, 65536)

            try:
                sock.bind(('::', port))
                break
            except OSError as e:
                if e.errno not in (errno.EADDRINUSE, errno.EADDRNOTAVAIL):
                    raise
        return port

    def retry(self):
        self.cancel_timeout()

        self._retry_count += 1
        if self._retry_count >= MAX_TRIES:
            self.fail_waiters(OSError(errno.ETIMEDOUT))
            if self.transport is not None:
                self.transport.abort()
            self.resolver().protocol_done(self)
            return

        if self.transport is not None:
            self.transport.abort()

        self.set_timeout(TIMEOUT)

        loop = asyncio.get_event_loop()

        self.server = next(self.server_selector)

        if (isinstance(self.server[0], ipaddress.IPv4Address)
            or (isinstance(self.server[0], str)
                and _ipv4_re.match(self.server[0]))):
            self.server = ('::ffff:' + str(self.server[0]), self.server[1])

        self.server = (str(self.server[0]), self.server[1])

        if not self.using_tcp:
            packet = build_dns_packet(self.uid, self.query,
                                      not self.recursive)

            if len(packet) > MAX_PACKET_SIZE:
                self.using_tcp = True
                del packet

        if self.using_tcp:
            self._buffer = b''

            sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
            sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
            self.bind_random_port(sock)
            f = asyncio.ensure_future(loop.create_connection(lambda: self,
                                                             host=self.server,
                                                             sock=sock))

            def callback(f):
                if f.cancelled():
                    self.cancel_waiters()
                    return

                exc = f.exception()
                if exc is not None:
                    self.fail_waiters(exc)
                    return

                packet = build_dns_packet(uid, query, not self.recursive)
                len_bytes = struct.pack(b'>H', len(packet))

                transport, _ = f.result()
                try:
                    transport.writelines(len_bytes, packet)
                except:
                    e = sys.exc_info()[0]
                    self.fail_waiters(e)

            f.add_done_callback(callback)
        else:
            self._buffer = None

            sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
            sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
            self.bind_random_port(sock)
            f = asyncio.ensure_future(loop.create_datagram_endpoint(lambda: self,
                                                                    sock=sock))

            def callback(f):
                if f.cancelled():
                    self.cancel_waiters()
                    return

                exc = f.exception()
                if exc is not None:
                    self.fail_waiters(exc)
                    return

                transport, _ = f.result()
                try:
                    transport.sendto(packet, addr=self.server)
                except:
                    e = sys.exc_info()[0]
                    self.fail_waiters(e)

            f.add_done_callback(callback)

    def connection_made(self, transport):
        self.transport = transport

    def datagram_received(self, data, addr):
        # Ignore responses from the wrong server
        if addr[:2] != self.server:
            return

        try:
            if self.process_packet(data):
                self.resolver().protocol_done(self)
            else:
                # This means we're re-querying using TCP, or recursing
                self._buffer = b''
        except:
            e = sys.exc_info()[0]
            self.fail_waiters(e)

    def data_received(self, data):
        self._buffer += data

        # Accumulate data until we have enough for the packet length
        if len(self._buffer) > 2:
            packet_len, = struct.unpack(b'>H', self._buffer[:2])

            # When we have enough for the entire packet, process it
            if len(self._buffer) >= 2 + packet_len:
                try:
                    done = self.process_packet(self._buffer[2:])
                except:
                    e = sys.exc_info()[0]
                    self.fail_waiters(e)
                finally:
                    self.transport.close()

                if not done:
                    # This means we're recursing
                    self._buffer = b''
                else:
                    self.resolver().protocol_done(self)

    def decode_rr(self, packet, ptr, rcode):
        name, ptr = decode_domain(packet, ptr)

        utype, uclss, ttl, rdlen = struct.unpack(b'>HHLH', packet[ptr:ptr+10])

        ptr += 10

        if utype == OPT:
            rcode |= (ttl >> 20) & 0xff0
            ptr += rdlen
            return (rcode, None, None)

        result = rr.RR.decode(name, utype, uclss, ttl, packet, ptr, rdlen)
        ptr += rdlen

        return (rcode, result, ptr)

    def process_packet(self, packet):
        self.cancel_timeout()

        # Decode the header
        uid, flags, qdcount, ancount, nscount, arcount \
            = struct.unpack(b'>HHHHHH', packet[:12])

        # Ignore packets that aren't query responses
        if (flags & QR) == 0:
            raise BadReply('Received a packet that isn\'t a response')

        if (flags & OPCODE_MASK) != QUERY:
            raise BadReply('Received a packet that isn\'t a query response')

        # Ignore packets with the wrong question count
        if qdcount != 1:
            raise BadReply('Too many questions in query response')

        # Ignore packets with the wrong uid
        if uid != self.uid:
            raise BadReply('Wrong UID in query response')

        # Decode the query
        domain, ptr = decode_domain(packet, 12)

        qtype, qclss = struct.unpack(b'>HH', packet[ptr:ptr+4])
        ptr += 4

        q = Query(domain, qtype, qclss)

        # Ignore packets containing the wrong query
        if q != self.query:
            raise BadReply('Query mismatch in query response - {} {}'.format(q, self.query))

        # Check the TC flag and re-query using TCP
        if (flags & TC) != 0:
            if self.using_tcp:
                raise BadReply('Query using TCP asks to re-do using TCP!')

            self.using_tcp = True
            self._retry_count = 0
            self.retry()

            return False

        # Find the rcode and decode the answer
        rcode = flags & RCODE_MASK

        answers = []
        authorities = []
        additional = []

        if rcode in (NOERROR, NXDOMAIN):
            for count, array in [(ancount, answers),
                                 (nscount, authorities),
                                 (arcount, additional)]:
                for n in range(0, count):
                    rcode, rr, ptr = self.decode_rr(packet, ptr, rcode)
                    if rr is not None:
                        array.append(rr)

            reply = Reply(flags,
                          rcode,
                          answers,
                          authorities,
                          additional)
        else:
            # Look for an OPT entry in the additional area to get the rcode

            # First skip the answer and authority parts
            for n in range(0, ancount + nscount):
                _, ptr = decode_domain(packet, ptr)
                utype, uclss, ttl, rdlen = struct.unpack(b'>HHLH',
                                                         packet[ptr:ptr+10])
                ptr += 10 + rdlen

            # Now read the additional entries
            for n in range(0, arcount):
                _, ptr = decode_domain(packet, ptr)
                utype, uclss, ttl, rdlen = struct.unpack(b'>HHLH',
                                                         packet[ptr:ptr+10])
                if utype == OPT:
                    rcode |= (ttl >> 20) & 0xff0

                ptr += 10 + rdlen

            reply = Reply(False, rcode, [], [], [])

        # Finally, check to see if we need to recurse
        if self.recursive \
           and (flags & (RA|RD)) != (RA|RD) \
           and rcode == NOERROR \
           and len(answers) == 0 and len(authorities) != 0:
            # Check to see if we have addresses in the additional section
            addrs = {}
            for addl in additional:
                if addl.rr_type not in (A, AAAA):
                    continue

                l = addrs.get(addl.name, None)
                if l is None:
                    l = list()
                    addrs[addl.name] = l
                l.append(addl.address)

            # Find the list of servers
            servers = []
            for authrr in authorities:
                if authrr.rr_type != NS:
                    continue
                addresses = addrs.get(authrr.host, None)
                got_v6 = False
                if self.prefer_ipv6:
                    for addr in addresses:
                        if isinstance(addr, ipaddress.IPv6Address):
                            servers.append((addr, 53))
                            got_v6 = True
                if not got_v6:
                    for addr in addresses:
                        if self.prefer_ipv6 is None \
                           or not isinstance(addr, ipaddress.IPv6Address):
                            servers.append((addr, 53))

            if len(servers):
                self.server_selector = RandomServer(servers)
                self.using_tcp = False
                self._retry_count = 0

                self.retry()

                return False

        self.fire_waiters(reply)

        return True

    def timed_out(self):
        self.fail_waiters(OSError(errno.ETIMEDOUT))

    def error_received(self, exc):
        self.fail_waiters(exc)

    def connection_lost(self, exc):
        if exc is not None:
            self.fail_waiters(exc)

    def add_waiter(self, future):
        self._waiters.append(future)

    def fire_waiters(self, reply):
        self.cancel_timeout()
        if self.should_cache and reply.rcode in (NOERROR, NXDOMAIN):
            self.resolver().cache_reply(self.query, reply)
        for f in self._waiters:
            f.set_result(reply)

    def cancel_waiters(self):
        self.cancel_timeout()
        for f in self._waiters:
            f.cancel()

    def fail_waiters(self, exc):
        self.cancel_timeout()
        for f in self._waiters:
            f.set_exception(exc)

class Resolver(object):

    def __init__(self):
        self._cache = {}
        self._queries = {}

    def close(self):
        """Cancel any in-progress queries."""
        if len(self._queries):
            for q,p in self._queries.items():
                p.cancel_waiters()

    def __del__(self):
        self.close()

    def check_domain(self, query):
        for label in query.name.split(b'.'):
            if len(label) > 63:
                raise ValueError('DNS label too long')

    def lookup(self, query, servers=None, prefer_ipv6=False, should_cache=True,
               recursive=False):
        """Perform a DNS query."""
        f = asyncio.Future()

        # First see if the result is in the cache - if so, return it
        r = self._cache.get(query, None)
        if r is not None:
            (expiry, reply) = r

            now = time.time()

            if now >= expiry:
                del _cache[query]
            else:
                ttl_ms = expiry - now
                ttl_s = (ttl_ms + 999) / 1000

                reply.update_ttls(ttl_s)

                f.set_result(reply)

                return f

        # Now, look to see if we're already doing this lookup; if we are,
        # add this future to the list waiting on the query.
        r = self._queries.get(query, None)
        if r is not None:
            r.add_waiter(f)
            return f

        # Finally, fire off a new query
        uid = _rng.randrange(0, 65536)
        self.check_domain(query)

        if servers is None:
            recursive = True
            if prefer_ipv6:
                servers = _ipv6_roots
            elif prefer_ipv6 is None:
                servers = _all_roots
            else:
                servers = _ipv4_roots

        if isinstance(servers, list):
            servers = RandomServer(servers)
        elif isinstance(servers, tuple):
            servers = RoundRobinServer([servers])

        protocol = DNSProtocol(self, query, servers, uid, False,
                               prefer_ipv6, should_cache, recursive)
        self._queries[query] = protocol
        protocol.add_waiter(f)
        protocol.retry()

        return f

    def protocol_done(self, protocol):
        del self._queries[protocol.query]

    def flush_cache(self):
        self._cache = {}

    def cache_reply(self, query, reply):
        now = time.time()
        has_ttl = False
        min_ttl = 0

        # Find the minimum ttl
        for collection in (reply.answers, reply.authorities, reply.additional):
            for rr in collection:
                ttl = rr.ttl
                if not has_ttl or min_ttl > ttl:
                    has_ttl = True
                    min_ttl = ttl

        # Can't cache things with no TTL or with TTL zero
        if not has_ttl:
            return

        if min_ttl > MAX_TTL:
            min_ttl = MAX_TTL

        self._cache[query] = (now + min_ttl, reply)



import re
import ipaddress
import time
import os

from .resolver import Resolver, RoundRobinServer, RandomServer

_space_re = re.compile(b'\\s+')

class SystemResolver(Resolver):

    def __init__(self):
        self._servers = None
        self._servers_timestamp = None

        super(SystemResolver, self).__init__()

    def read_servers(self):
        servers = []
        with open('/etc/resolv.conf', 'rb') as f:
            for line in f:
                line = line.strip()
                if not line or line.startswith(b'#'):
                    continue

                fields = _space_re.split(line)
                if len(fields) < 2:
                    continue

                if fields[0] != b'nameserver':
                    continue

                try:
                    addr = ipaddress.ip_address(fields[1].decode('ascii'))
                except ValueError:
                    continue

                servers.append((addr, 53))

        self._servers = RoundRobinServer(servers)

    def maybe_read_servers(self):
        now = time.time()
        if self._servers_timestamp is not None \
           and now - self._servers_timestamp < 30:
            return

        s = os.stat('/etc/resolv.conf')
        if self._servers_timestamp is None \
           or s.st_mtime > self._servers_timestamp:
            self._servers_timestamp = s.st_mtime
            self.read_servers()

    def lookup(self, query, prefer_ipv6=False,
               should_cache=True, recursive=False):
        self.maybe_read_servers()
        return super(SystemResolver, self).lookup(query, self._servers,
                                                  prefer_ipv6, should_cache,
                                                  recursive)



import ipaddress
import struct

from . import constants, utils
from .utils import rrclass_to_string, rrtype_to_string, escape_string

_rr_registry = {}

class RR(object):
    def __init__(self, name, rr_type, rr_class, ttl):
        self.name = name
        self.rr_type = rr_type
        self.rr_class = rr_class
        self.ttl = ttl

    def __str__(self):
        return '{}\t{}\t{}\t{}'.format(self.name,
                                       self.ttl,
                                       rrclass_to_string(self.rr_class),
                                       rrtype_to_string(self.rr_type))

    @staticmethod
    def register(rr_type, rr_class, pyclass):
        """Register a new RR type."""
        _rr_registry[(rr_class, rr_type)] = pyclass

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        pyclass = _rr_registry.get((rr_class, rr_type),
                                   _rr_registry.get((constants.ANY, rr_type),
                                                    Unknown))
        return pyclass.decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen)

def rr(rr_type, rr_class):
    def inner(clss):
        RR.register(rr_type, rr_class, clss)
        return clss
    return inner

@rr(constants.A, constants.IN)
class A(RR):
    def __init__(self, name, ttl, address):
        super(A, self).__init__(name, constants.A, constants.IN, ttl)
        self.address = address

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}'.format(escape_string(self.name),
                                           self.ttl,
                                           rrclass_to_string(self.rr_class),
                                           rrtype_to_string(self.rr_type),
                                           self.address)

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        addr = ipaddress.IPv4Address(packet[ptr:ptr+4])
        return A(name, ttl, addr)

@rr(constants.AAAA, constants.IN)
class AAAA(RR):
    def __init__(self, name, ttl, address):
        super(AAAA, self).__init__(name, constants.AAAA, constants.IN, ttl)
        self.address = address

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}'.format(escape_string(self.name),
                                           self.ttl,
                                           rrclass_to_string(self.rr_class),
                                           rrtype_to_string(self.rr_type),
                                           self.address)

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        addr = ipaddress.IPv6Address(packet[ptr:ptr+16])
        return AAAA(name, ttl, addr)

@rr(constants.CNAME, constants.IN)
class CNAME(RR):
    def __init__(self, name, ttl, cname):
        super(CNAME, self).__init__(name, constants.CNAME, constants.IN, ttl)
        self.cname = cname

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}'.format(escape_string(self.name),
                                           self.ttl,
                                           rrclass_to_string(self.rr_class),
                                           rrtype_to_string(self.rr_type),
                                           escape_string(self.cname))

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        cname, _ = utils.decode_domain(packet, ptr)
        return CNAME(name, ttl, cname)

@rr(constants.HINFO, constants.IN)
class HINFO(RR):
    def __init__(self, name, ttl, cpu, os):
        super(HINFO, self).__init__(name, constants.HINFO, constants.IN, ttl)
        self.cpu = cpu
        self.os = os

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}\t{}'.format(escape_string(self.name),
                                               self.ttl,
                                               rrclass_to_string(self.rr_class),
                                               rrtype_to_string(self.rr_type),
                                               escape_string(self.cpu),
                                               escape_string(self.os))

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        cpu, ptr = utils.decode_pascal_string(packet, ptr)
        os, ptr = utils.decode_pascal_string(packet, ptr)
        return HINFO(name, ttl, cpu, os)

@rr(constants.MB, constants.IN)
class MB(RR):
    def __init__(self, name, ttl, host):
        super(MB, self).__init__(name, constants.MB, constants.IN, ttl)
        self.host = host

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}'.format(escape_string(self.name),
                                           self.ttl,
                                           rrclass_to_string(self.rr_class),
                                           rrtype_to_string(self.rr_type),
                                           escape_string(self.host))

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        host, _ = utils.decode_domain(packet, ptr)
        return MB(name, ttl, host)

@rr(constants.MF, constants.IN)
class MF(RR):
    def __init__(self, name, ttl, host):
        super(MF, self).__init__(name, constants.MF, constants.IN, ttl)
        self.host = host

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}'.format(escape_string(self.name),
                                           self.ttl,
                                           rrclass_to_string(self.rr_class),
                                           rrtype_to_string(self.rr_type),
                                           escape_string(self.host))

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        host, _ = utils.decode_domain(packet, ptr)
        return MF(name, ttl, host)

@rr(constants.MG, constants.IN)
class MG(RR):
    def __init__(self, name, ttl, mailbox):
        super(MG, self).__init__(name, constants.MG, constants.IN, ttl)
        self.mailbox = mailbox

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}'.format(escape_string(self.name),
                                           self.ttl,
                                           rrclass_to_string(self.rr_class),
                                           rrtype_to_string(self.rr_type),
                                           escape_string(self.mailbox))

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        mailbox, _ = utils.decode_domain(packet, ptr)
        return MG(name, ttl, mailbox)

@rr(constants.MINFO, constants.IN)
class MINFO(RR):
    def __init__(self, name, ttl, rmailbox, emailbox):
        super(MINFO, self).__init__(name, constants.MINFO, constants.IN, ttl)
        self.rmailbox = rmailbox
        self.emailbox = emailbox

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}\t{}'.format(escape_string(self.name),
                                               self.ttl,
                                               rrclass_to_string(self.rr_class),
                                               rrtype_to_string(self.rr_type),
                                               escape_string(self.rmailbox),
                                               escape_string(self.emailbox))

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        rmailbox, ptr = utils.decode_domain(packet, ptr)
        emailbox, ptr = utils.decode_domain(packet, ptr)
        return MINFO(name, ttl, rmailbox, emailbox)

@rr(constants.MR, constants.IN)
class MR(RR):
    def __init__(self, name, rr_class, ttl, mailbox):
        super(MR, self).__init__(name, constants.MR, rr_class, ttl)
        self.mailbox = mailbox

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}'.format(escape_string(self.name),
                                           self.ttl,
                                           rrclass_to_string(self.rr_class),
                                           rrtype_to_string(self.rr_type),
                                           escape_string(self.mailbox))

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        mailbox, _ = utils.decode_domain(packet, ptr)
        return MG(name, ttl, mailbox)

@rr(constants.MX, constants.IN)
class MX(RR):
    def __init__(self, name, ttl, preference, exchange):
        super(MX, self).__init__(name, constants.MX, constants.IN, ttl)
        self.preference = preference
        self.exchange = exchange

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}\t{}'.format(escape_string(self.name),
                                               self.ttl,
                                               rrclass_to_string(self.rr_class),
                                               rrtype_to_string(self.rr_type),
                                               self.preference,
                                               escape_string(self.exchange))

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        preference, = struct.unpack(b'>H', packet[ptr:ptr+2])
        exchange, _ = utils.decode_domain(packet, ptr+2)
        return MX(name, ttl, preference, exchange)

@rr(constants.NUL, constants.IN)
class NUL(RR):
    def __init__(self, name, ttl, data):
        super(NULL, self).__init__(name, constants.NUL, constants.IN, ttl)
        self.data = data

    def __str__(self):
        return '{}\t{}\t{}\t{}\t({})'.format(escape_string(self.name),
                                             self.ttl,
                                             rrclass_to_string(self.rr_class),
                                             rrtype_to_string(self.rr_type),
                                             utils.base64(self.data))

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        return NUL(name, ttl, packet[ptr:ptr+rdlen])

@rr(constants.NS, constants.IN)
class NS(RR):
    def __init__(self, name, ttl, host):
        super(NS, self).__init__(name, constants.NS, constants.IN, ttl)
        self.host = host

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}'.format(escape_string(self.name),
                                           self.ttl,
                                           rrclass_to_string(self.rr_class),
                                           rrtype_to_string(self.rr_type),
                                           escape_string(self.host))

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        host, _ = utils.decode_domain(packet, ptr)
        return NS(name, ttl, host)

@rr(constants.PTR, constants.IN)
class PTR(RR):
    def __init__(self, name, ttl, dname):
        super(PTR, self).__init__(name, constants.PTR, constants.IN, ttl)
        self.dname = dname

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}'.format(escape_string(self.name),
                                           self.ttl,
                                           rrclass_to_string(self.rr_class),
                                           rrtype_to_string(self.rr_type),
                                           escape_string(self.dname))

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        dname, _ = utils.decode_domain(packet, ptr)
        return PTR(name, ttl, dname)

@rr(constants.SOA, constants.IN)
class SOA(RR):
    def __init__(self, name, ttl, mname, rname, serial,
                 refresh, retry, expire, minimum):
        super(SOA, self).__init__(name, constants.SOA, constants.IN, ttl)
        self.mname = mname
        self.rname = rname
        self.serial = serial
        self.refresh = refresh
        self.retry = retry
        self.expire = expire
        self.minimum = minimum

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}'.format(
            escape_string(self.name),
            self.ttl,
            rrclass_to_string(self.rr_class),
            rrtype_to_string(self.rr_type),
            escape_string(self.mname), escape_string(self.rname),
            self.serial, self.refresh,
            self.expire, self.minimum)

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        mname, ptr = utils.decode_domain(packet, ptr)
        rname, ptr = utils.decode_domain(packet, ptr)
        serial, refresh, retry, expire, minimum = struct.unpack('>LLLLL',
                                                                packet[ptr:ptr+20])
        return SOA(name, ttl, mname, rname, serial,
                   refresh, retry, expire, minimum)

@rr(constants.TXT, constants.IN)
class TXT(RR):
    def __init__(self, name, ttl, text):
        super(TXT, self).__init__(name, constants.TXT, constants.IN, ttl)
        self.text = text

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}'.format(escape_string(self.name),
                                           self.ttl,
                                           rrclass_to_string(self.rr_class),
                                           rrtype_to_string(self.rr_type),
                                           escape_string(self.txt))

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        text = []
        while ptr < len(packet):
            chunk, ptr = utils.decode_pascal_string(packet, ptr)
            text.append(chunk)
        text = ''.join(text)
        return TXT(name, ttl, text)

@rr(constants.WKS, constants.IN)
class WKS(RR):
    def __init__(self, name, ttl, address, protocol, bitmap):
        super(WKS, self).__init__(name, constants.WKS, constants.IN, ttl)
        self.address = address
        self.protocol = protocol
        self.bitmap = bitmap

    def __str__(self):
        bits = []
        base = 0
        for b in self.bitmap:
            for o in range(0, 8):
                bit = 0x80 >> o
                if b & bit:
                    bits.push_back(base + o)
            base += 8

        return '{}\t{}\t{}\t{}\t{}\t{}\t({})'.format(
            escape_string(self.name),
            self.ttl,
            rrclass_to_string(self.rr_class),
            rrtype_to_string(self.rr_type),
            self.address,
            self.protocol,
            ', '.join(str(bit) for bit in bits))

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        addr = ipaddress.IPv4Address(packet[ptr:ptr+4])
        protocol = packet[ptr+4]
        bitmap = packet[ptr+5:]
        return WKS(name, ttl, addr, protocol, bitmap)

class Unknown(RR):
    def __init__(self, name, rr_type, rr_class, ttl, data):
        super(Unknown, self).__init__(name, rr_type, rr_class, ttl)
        self.data = data

    def __str__(self):
        return '{}\t{}\t{}\t{}\t({})'.format(escape_string(self.name),
                                             self.ttl,
                                             rrclass_to_string(self.rr_class),
                                             rrtype_to_string(self.rr_type),
                                             utils.base64(self.data))

    @staticmethod
    def decode(name, rr_type, rr_class, ttl, packet, ptr, rdlen):
        return Unknown(name, rr_type, rr_class, ttl, packet[ptr:ptr+rdlen])



#
# A Windows-specific SystemResolver implementation that gets the DNS server
# list from the Windows API
#
import ipaddress
import time

from ctypes import *

from .resolver import Resolver, RoundRobinServer, RandomServer

iphlpapi = windll.iphlpapi

UINT = c_uint
DWORD = c_uint32
ULONG = c_ulong
PULONG = POINTER(c_ulong)

class IP_ADDRESS_STRING(Structure):
    _fields_ = [('String', c_char * (4 * 4))]

IP_MASK_STRING = IP_ADDRESS_STRING

class IP_ADDR_STRING(Structure):
    pass

IP_ADDR_STRING._fields_ = [('Next', POINTER(IP_ADDR_STRING)),
                           ('IpAddress', IP_ADDRESS_STRING),
                           ('IpMask', IP_MASK_STRING),
                           ('Context', DWORD)]

PIP_ADDR_STRING = POINTER(IP_ADDR_STRING)

MAX_HOSTNAME_LEN = 128
MAX_DOMAIN_NAME_LEN = 128
MAX_SCOPE_ID_LEN = 256
class FIXED_INFO(Structure):
    _fields_ = [('HostName', c_char * (MAX_HOSTNAME_LEN + 4)),
                ('DomainName', c_char * (MAX_DOMAIN_NAME_LEN + 4)),
                ('CurrentDnsServer', PIP_ADDR_STRING),
                ('DnsServerList', IP_ADDR_STRING),
                ('NodeType', UINT),
                ('ScopeId', c_char * (MAX_SCOPE_ID_LEN + 4)),
                ('EnableRouting', UINT),
                ('EnableProxy', UINT),
                ('EnableDns', UINT)]
PFIXED_INFO = POINTER(FIXED_INFO)

ERROR_SUCCESS = 0
ERROR_BUFFER_OVERFLOW = 111

GetNetworkParams = iphlpapi.GetNetworkParams
GetNetworkParams.restype = DWORD
GetNetworkParams.args = [PFIXED_INFO, PULONG]

class SystemResolver(Resolver):

    def __init__(self):
        self._servers = None
        self._servers_timestamp = None

        super(SystemResolver, self).__init__()

    def read_servers(self):
        servers = []

        outLen = ULONG(0)
        ret = GetNetworkParams(None, byref(outLen))
        if ret != ERROR_BUFFER_OVERFLOW:
            raise Exception('Unexpected error {:08x}'.format(ret))
        buf = create_string_buffer(outLen.value)
        pfi = cast(buf, PFIXED_INFO)
        GetNetworkParams(pfi, byref(outLen))
        fi = pfi.contents

        try:
            saddr = fi.DnsServerList.IpAddress.String
            addr = ipaddress.ip_address(saddr.decode('ascii'))
            servers.append((addr, 53))
        except ValueError:
            pass

        pipa = fi.DnsServerList.Next
        while pipa:
            try:
                saddr = pipa.contents.IPAddress.String
                addr = ipaddress.ip_address(saddr.decode('ascii'))
                servers.append((addr, 53))
            except ValueError:
                pass
            pipa = pipa.contents.Next

        self._servers = RoundRobinServer(servers)

    def maybe_read_servers(self):
        now = time.time()
        if self._servers_timestamp is not None \
           and now - self._servers_timestamp < 30:
            return

        self.read_servers()

    def lookup(self, query, prefer_ipv6=False,
               should_cache=True, recursive=False):
        self.maybe_read_servers()
        return super(SystemResolver, self).lookup(query, self._servers,
                                                  prefer_ipv6, should_cache,
                                                  recursive)



from .constants import *

from . import rr, dnssec

from .resolver import Query, Reply, Resolver
from .mdns import MulticastResolver
from .hosts import HostsResolver
from .system import SystemResolver
from .smart import SmartResolver



import ipaddress
import re
import os
import time
import asyncio
import sys

from . import rr

from .constants import *
from .wireformat import *
from .resolver import Query, Reply

if sys.platform == 'win32':
    _hosts_path = 'C:/Windows/System32/drivers/etc/hosts'
else:
    _hosts_path = '/etc/hosts'

_space_re = re.compile(b'\\s+')
class HostsResolver(object):
    """Resolves queries using the entries in /etc/hosts."""

    def __init__(self):
        self._hosts_timestamp = None
        self._hosts = {}
        self._addrs = {}

    def check_domain(self, query):
        for label in query.name.split(b'.'):
            if len(label) > 63:
                raise ValueError('DNS label too long')

    def read_hosts(self):
        self._hosts = {}
        self._addrs = {}
        with open(_hosts_path, 'rb') as f:
            for line in f:
                line = line.strip()
                if not line or line.startswith(b'#'):
                    continue
                fields = _space_re.split(line)
                if len(fields) < 2:
                    continue
                name = fields[1]
                try:
                    addr = ipaddress.ip_address(fields[0].decode('ascii'))
                except ValueError:
                    continue

                name = name.lower()
                l = self._hosts.get(name, None)
                if l is None:
                    l = []
                    self._hosts[name] = l
                l.append(addr)
                self._addrs[addr] = name

        # Always add localhost if it isn't there already
        lhv4 = ipaddress.ip_address('127.0.0.1')
        lhv6 = ipaddress.ip_address('::1')
        self._hosts[b'localhost'] = [ lhv4, lhv6 ]
        self._addrs[lhv4] = b'localhost'
        self._addrs[lhv6] = b'localhost'

    def maybe_read_hosts(self):
        now = time.time()
        if self._hosts_timestamp is not None \
           and now - self._hosts_timestamp < 30:
            return

        s = os.stat(_hosts_path)
        if self._hosts_timestamp is None \
           or s.st_mtime > self._hosts_timestamp:
            self._hosts_timestamp = s.st_mtime
            self.read_hosts()

    def lookup(self, query):
        f = asyncio.Future()

        self.maybe_read_hosts()

        if query.name.endswith(b'.in-addr.arpa'):
            addr = b'.'.join(query.name[:-13].split(b'.')[::-1])
            try:
                addr = ipaddress.ip_address(addr.decode('ascii'))
            except ValueError:
                addr = None

            hostname = self._addrs.get(addr, None)
            if addr is None or hostname is None:
                reply = Reply(QR, NXDOMAIN, [], [], [])
            elif query.q_class in (IN, ANY) and query.q_type in (PTR, ANY):
                answers = [rr.PTR(query.name, 30, hostname)]
                host = self._hosts[hostname]
                additional = []
                for addr in host:
                    if isinstance(addr, ipaddress.IPv4Address):
                        additional.append(rr.A(hostname, 30, addr))
                    elif isinstance(addr, ipaddress.IPv6Address):
                        additional.append(rr.AAAA(hostname, 30, addr))
                reply = Reply(QR, NOERROR, answers, [], additional)
            else:
                reply = Reply(QR, NOERROR, [], [], [])
        elif query.name.endswith(b'.ip6.arpa'):
            addr = b''.join(query.name[:-9].split(b'.')[::-1])
            addr = b':'.join([addr[n:n+4] for n in range(0,len(addr),4)])
            try:
                addr = ipaddress.ip_address(addr.decode('ascii'))
            except ValueError:
                addr = None

            hostname = self._addrs.get(addr, None)
            if addr is None or hostname is None:
                reply = Reply(QR, NXDOMAIN, [], [], [])
            elif query.q_class in (IN, ANY) and query.q_type in (PTR, ANY):
                answers = [rr.PTR(query.name, 30, hostname)]
                host = self._hosts[hostname]
                additional = []
                for addr in host:
                    if isinstance(addr, ipaddress.IPv4Address):
                        additional.append(rr.A(hostname, 30, addr))
                    elif isinstance(addr, ipaddress.IPv6Address):
                        additional.append(rr.AAAA(hostname, 30, addr))
                reply = Reply(QR, NOERROR, answers, [], additional)
            else:
                reply = Reply(QR, NOERROR, [], [], [])
        elif query.q_class in (IN, ANY) and query.q_type == A:
            host = self._hosts.get(query.name, None)
            if host is None:
                reply = Reply(QR, NXDOMAIN, [], [], [])
            else:
                answers = []
                additional = []
                for addr in host:
                    if isinstance(addr, ipaddress.IPv4Address):
                        answers.append(rr.A(query.name, 30, addr))
                    elif isinstance(addr, ipaddress.IPv6Address):
                        additional.append(rr.AAAA(query.name, 30, addr))
                reply = Reply(QR, NOERROR, answers, [], additional)
        elif query.q_class in (IN, ANY) and query.q_type == AAAA:
            host = self._hosts.get(query.name, None)
            if host is None:
                reply = Reply(QR, NXDOMAIN, [], [], [])
            else:
                answers = []
                additional = []
                for addr in host:
                    if isinstance(addr, ipaddress.IPv4Address):
                        additional.append(rr.A(query.name, 30, addr))
                    elif isinstance(addr, ipaddress.IPv6Address):
                        answers.append(rr.AAAA(query.name, 30, addr))
                reply = Reply(QR, NOERROR, answers, [], additional)
        elif query.q_class in (IN, ANY) and query.q_type == ANY:
            host = self._hosts.get(query.name, None)
            if host is None:
                reply = Reply(QR, NXDOMAIN, [], [], [])
            else:
                answers = []
                for addr in host:
                    if isinstance(addr, ipaddress.IPv4Address):
                        answers.append(rr.A(query.name, 30, addr))
                    elif isinstance(addr, ipaddress.IPv6Address):
                        answers.append(rr.AAAA(query.name, 30, addr))
                reply = Reply(QR, NOERROR, answers, [], [])
        else:
            if query.name in self._hosts:
                reply = Reply(QR, NOERROR, [], [], [])
            else:
                reply = Reply(QR, NXDOMAIN, [], [], [])

        f.set_result(reply)

        return f



# -- rcodes

# RFC 1035
NOERROR  = 0
FORMERR  = 1
SERVFAIL = 2
NXDOMAIN = 3
NOTIMP   = 4
REFUSED  = 5

# RFC 2136
YXDOMAIN = 6
YXRRSET  = 7
NXRRSET  = 8
NOTAUTH  = 9
NOTZONE  = 10

# RFC 2671
BADVERS  = 16

# RFC 2845
BADSIG   = 16
BADKEY   = 17
BADTIME  = 18

# RFC 2930
BADMODE  = 19
BADNAME  = 20
BADALG   = 21

# -- RR types

A          = 1
NS         = 2
MD         = 3
MF         = 4
CNAME      = 5
SOA        = 6
MB         = 7
MG         = 8
MR         = 9
NUL        = 10
WKS        = 11
PTR        = 12
HINFO      = 13
MINFO      = 14
MX         = 15
TXT        = 16
RP         = 17
AFSDB      = 18
X25        = 19
ISDN       = 20
RT         = 21
NSAP       = 22
NSAPPTR    = 23
SIG        = 24
KEY        = 25
PX         = 26
GPOS       = 27
AAAA       = 28
LOC        = 29
NXT        = 30
EID        = 31
NIMLOC     = 32
SRV        = 33
ATMA       = 34
NAPTR      = 35
KX         = 36
CERT       = 37
A6         = 38
DNAME      = 39
SINK       = 40
OPT        = 41
APL        = 42
DS         = 43
SSHFP      = 44
IPSECKEY   = 45
RRSIG      = 46
NSEC       = 47
DNSKEY     = 48
DHCID      = 49
NSEC3      = 50
NSEC3PARAM = 51
TLSA       = 52
HIP        = 55
CDS        = 59
CDNSKEY    = 60
OPENPGPKEY = 61

SPF        = 99
UINFO      = 100
UID        = 101
GID        = 102
UNSPEC     = 103

TKEY       = 249
TSIG       = 250
IXFR       = 251
AXFR       = 252
MAILB      = 253
MAILA      = 254
ANY        = 255
URI        = 256

CAA        = 257
TA         = 32768
DLV        = 32769

# -- RR class

IN   = 1
CS   = 2
CH   = 3
HS   = 4

NONE = 254
#ANY  = 255  # But we've already defined this above



import asyncio
import socket
import weakref
import errno
import random
import struct
import time

from . import rr
from .constants import *
from .utils import decode_domain, decode_pascal_string, rcode_to_string, \
    build_dns_packet, MAX_PACKET_SIZE
from .wireformat import *
from .timeout import Timeout
from .resolver import Query, Reply

TIMEOUT   = 5
MAX_TRIES = 5
MAX_TTL   = 3 * 3600

class MDNSQuery(object):
    """Represents an active MDNS query."""

    def __init__(self, resolver, protocol, query, use_ipv6, unicast_reply):
        self.resolver = weakref.ref(resolver)
        self.protocol = weakref.ref(protocol)
        self.query = query
        self.use_ipv6 = use_ipv6
        self.unicast_reply = unicast_reply

        self._retry_count = 0
        self._waiters = []
        self._timeout = None

    def set_timeout(self, timeout):
        self.cancel_timeout()
        self._timeout = Timeout(timeout, self.timed_out)

    def cancel_timeout(self):
        if self._timeout is not None:
            self._timeout.cancel()
            self._timeout = None

    def timed_out(self):
        self.retry()

    def add_waiter(self, f):
        self._waiters.append(f)

    def fire_waiters(self, reply):
        self.cancel_timeout()
        for f in self._waiters:
            f.set_result(reply)
        self.resolver().query_done(self)

    def cancel_waiters(self):
        self.cancel_timeout()
        for f in self._waiters:
            f.cancel()
        self.resolver().query_done(self)

    def fail_waiters(self, exc):
        self.cancel_timeout()
        for f in self._waiters:
            f.set_exception(exc)
        self.resolver().query_done(self)

    def retry(self):
        self.cancel_timeout()

        self._retry_count += 1
        if self._retry_count >= MAX_TRIES:
            self.fire_waiters(Reply(QR, NXDOMAIN, [], [], []))
            return

        self.set_timeout(TIMEOUT)

        packet = build_dns_packet(0, self.query, False, True)

        if len(packet) > MAX_PACKET_SIZE:
            del packet
            self.fail_waiters(OSError(errno.ETOOBIG))

        self.protocol().send_packet(packet, self.use_ipv6)

_rng = random.SystemRandom()

class MulticastResolver(object):
    """Resolves queries using Multicast DNS (aka MDNS, aka Bonjour)."""

    def __init__(self):
        self._cache = {}
        self._queries = {}
        self._queue = []
        self._socket = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
        self._socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)

        # Bind this socket - we need to do this on Windows or we get an
        # exception inside asyncio because it tries to use getsockname().
        while True:
            port = _rng.randrange(1024, 65536)
            try:
                self._socket.bind(('::', port))
                break
            except OSError as e:
                if e.errno not in (errno.EADDRINUSE, EADDRNOTAVAIL):
                    raise
        
        ip_mreq = b'\xe0\x00\x00\xfb\x00\x00\x00\x00'
        self._socket.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP,
                                ip_mreq)
        self.transport = None
        loop = asyncio.get_event_loop()
        f = asyncio.ensure_future(loop.create_datagram_endpoint(lambda: self,
                                                                sock=self._socket))

        def callback(f):
            if f.cancelled():
                self.cancel_all_waiters()
                return

            exc = f.exception()
            if exc is not None:
                self.fail_all_waiters(exc)
                return

        f.add_done_callback(callback)

    def close(self):
        self.cancel_all_waiters()
        self.transport.abort()

    def __del__(self):
        # On Windows, we can get a (harmless) RuntimeError here from trying
        # to abort the transport after the event loop has finished.
        try:
            self.close()
        except RuntimeError:
            pass
    
    def connection_made(self, transport):
        self.transport = transport
        for item in self._queue:
            transport.sendto(item[0], addr=item[1])
        self._queue = []

    def connection_lost(self, exc):
        if exc is not None:
            self.fail_all_waiters(exc)

    def datagram_received(self, data, addr):
        try:
            self.process_packet(data)
        except:
            pass

    def cancel_all_waiters(self):
        for q in self._queries.values():
            q.cancel_waiters()
        self._queries = {}

    def fail_all_waiters(self, exc):
        for q in self._queries.values():
            q.fail_waiters(exc)
        self._queries = {}

    def check_domain(self, query):
        for label in query.name.split(b'.'):
            if len(label) > 63:
                raise ValueError('DNS label too long')

    def lookup(self, query, use_ipv6=False, unicast_reply=False):
        """Perform an MDNS query."""
        f = asyncio.Future()

        # First see if the result is in the cache - if so, return it
        r = self._cache.get(query, None)
        if r is not None:
            (expiry, reply) = r

            now = time.time()

            if now >= expiry:
                del _cache[query]
            else:
                ttl_ms = expiry - now
                ttl_s = (ttl_ms + 999) / 1000

                reply.update_ttls(ttl_s)

                f.set_result(reply)

                return f

        # Add this future to the list waiting on this query
        r = self._queries.get(query, None)
        if r:
            r.add_waiter(f)
            return f

        # Finally, fire off a new query
        self.check_domain(query)

        new_query = MDNSQuery(self, self, query, use_ipv6, unicast_reply)
        self._queries[query] = new_query
        new_query.add_waiter(f)
        new_query.retry()

        return f

    def query_done(self, query):
        del self._queries[query.query]

    def flush_cache(self):
        self._cache = {}

    def cache_reply(self, query, reply):
        now = time.time()
        has_ttl = False
        min_ttl = 0

        # Find the minimum ttl
        for collection in (reply.answers, reply.authorities, reply.additional):
            for rr in collection:
                ttl = rr.ttl
                if not has_ttl or min_ttl > ttl:
                    has_ttl = True
                    min_ttl = ttl

        # Can't cache things with no TTL or with TTL zero
        if not has_ttl:
            return

        if min_ttl > MAX_TTL:
            min_ttl = MAX_TTL

        self._cache[query] = (now + min_ttl, reply)

    def decode_rr(self, packet, ptr, rcode):
        name, ptr = decode_domain(packet, ptr)

        utype, uclss, ttl, rdlen = struct.unpack(b'>HHLH', packet[ptr:ptr+10])

        ptr += 10

        if utype == OPT:
            rcode |= (ttl >> 20) & 0xff0
            ptr += rdlen
            return (rcode, None, None)

        result = rr.RR.decode(name, utype, uclss, ttl, packet, ptr, rdlen)
        ptr += rdlen

        return (rcode, result, ptr)

    def send_packet(self, packet, use_ipv6=False):
        if use_ipv6:
            addr = ('ff02::fb', 5353)
        else:
            addr = ('224.0.0.251', 5353)

        if self.transport is None:
            self._queue.append((packet, addr))
        else:
            self.transport.sendto(packet, addr=addr)

    def process_packet(self, packet):
        # Decode the header
        uid, flags, qdcount, ancount, nscount, arcount \
            = struct.unpack(b'>HHHHHH', packet[:12])

        # Ignore things other than queries
        if (flags & OPCODE_MASK) != QUERY:
            return

        # Find the rcode
        rcode = flags & RCODE_MASK

        # Ignore anything other than NOERROR
        if rcode != NOERROR:
            return

        # Decode the query
        domain, ptr = decode_domain(packet, 12)

        qtype, qclss = struct.unpack(b'>HH', packet[ptr:ptr+4])
        ptr += 4

        q = Query(domain, qtype, qclss)

        # Look for the query
        found_query = None
        found_mq = None
        for fq,p in self._queries.items():
            if fq == q:
                print('Found')
                found_query = fq
                found_mq = p
                break

        # Decode the reply
        answers = []
        authorities = []
        additional = []

        for count, array in [(ancount, answers),
                             (nscount, authorities),
                             (arcount, additional)]:
            for n in range(0, count):
                rcode, rr, ptr = self.decode_rr(packet, ptr, rcode)
                if rr is not None:
                    array.append(rr)

            reply = Reply(flags,
                          rcode,
                          answers,
                          authorities,
                          additional)

        # For each answer, see if we can find a query that matches
        for answer in answers:
            for fq,p in self._queries.items():
                if fq.name.lower() == answer.name.lower() \
                   and (fq.q_type == ANY or fq.q_type == answer.rr_type) \
                   and (fq.q_class == ANY or fq.q_class == answer.rr_class):
                    self.cache_reply(fq, reply)
                    p.fire_waiters(reply)



import base64 as b64
import struct

from .wireformat import *
from . import constants

MAX_PACKET_SIZE = 4000

_rcode_strings = [ 'No error',
                   'Format error',
                   'Server failure',
                   'Non-existent domain',
                   'Not implemented',
                   'Query refused',
                   'Name exists when it should not',
                   'RR set exists when it should not',
                   'RR set that should exist does not',
                   'Server not authoritative for zone',
                   'Name not contained in zone',
                   None,
                   None,
                   None,
                   None,
                   None,
                   'Bad OPT version OR TSIG signature failure',
                   'Key not recognized',
                   'Signature out of time window',
                   'Bad TKEY mode',
                   'Duplicate key name',
                   'Algorithm not supported' ]

_rrtype_strings = [ None, 'A', 'NS', 'MD', 'MF', 'CNAME', 'SOA', 'MB', 'MG', 'MR',
                    'NUL', 'WKS', 'PTR', 'HINFO', 'MINFO', 'MX', 'TXT', 'RP',
                    'AFSDB', 'X25', 'ISDN', 'RT', 'NSAP', 'NSAPPTR', 'SIG',
                    'KEY', 'PX', 'GPOS',
                    'AAAA', 'LOC', 'NXT', 'EID', 'NIMLOC', 'SRV', 'ATMA', 'NAPTR',
                    'KX', 'CERT', 'A6', 'DNAME', 'SINK', 'OPT', 'APL', 'DS',
                    'SSHFP', 'IPSECKEY', 'RRSIG', 'NSEC', 'DNSKEY', 'DHCID',
                    'NSEC3', 'NSEC3PARAM', 'TLSA', None, None, 'HIP', None,
                    None, None, 'CDS', 'CDNSKEY', 'OPENPGPKEY' ]
_rrtype_extras = { 99: 'SPF', 100: 'UINFO', 101: 'UID', 102: 'GID', 103: 'UNSPEC',
                   249: 'TKEY', 250: 'TSIG', 251: 'IXFR', 252: 'AXFR',
                   253: 'MAILB', 254: 'MAILA', 255: 'ANY', 256: 'URI',
                   257: 'CAA', 32768: 'TA', 32769: 'DLV' }

_rrclass_strings = [ None, 'IN', 'CS', 'CH', 'HS' ]

def escape_string(byte_string):
    try:
        ustr = byte_string.decode('ascii')
        return ustr
    except UnicodeError:
        ustr = byte_string.decode('ascii', 'backslashreplace').sub('"', '\\"')
        return '"{}"'.format(ustr)

def base64(byte_string):
    return b64.b64encode(byte_string).decode('ascii')

def rcode_to_string(rcode):
    """Convert an RCODE to a string"""
    try:
        s = _rcode_strings[rcode]
    except KeyError:
        s = None
    if s is None:
        s = 'Unknown ({})'.format(rcode)
    return s

def rrtype_to_string(rrt):
    """Convert an RR type to a string"""
    try:
        s = _rrtype_strings[rrt]
    except KeyError:
        s = _rrtype_extras.get(rrt, None)

    if s is None:
        s = 'TYPE{}'.format(rrt)

    return s

def rrclass_to_string(rrt):
    """Convert an RR class to a string"""
    try:
        s = _rrclass_strings[rrt]
    except KeyError:
        s = None

    if s is None:
        if rrt == NONE:
            s = 'NONE'
        elif rrt == ANY:
            s = 'ANY'
        else:
            s = 'CLASS{}'.format(rrt)

    return s

def decode_domain(packet, ptr):
    result = []
    saved = False
    saved_ptr = None

    while True:
        length = packet[ptr]
        ptr += 1

        if not length:
            break

        if length < 64:
            result.append(packet[ptr:ptr+length])
            ptr += length
        elif (length & 0xc0) == 0xc0:
            low = packet[ptr]
            ptr += 1

            offset = ((length & 0x3f) << 8) | low

            if offset > len(packet):
                raise ValueError('Bad reply to DNS query')

            if not saved:
                saved = True
                saved_ptr = ptr

            ptr = offset

    if saved:
        ptr = saved_ptr

    return (b'.'.join(result), ptr)

def decode_pascal_string(packet, ptr):
    slen = packet[ptr]
    ptr += 1
    s = packet[ptr:ptr+slen]
    ptr += slen
    return (s.decode('utf8'), ptr)

def build_dns_packet(uid, query, wants_recursion=False, unicast=False):
    flags = QUERY
    if wants_recursion:
        flags |= RD
    header = struct.pack(b'>HHHHHH', uid, flags, 1, 0, 0, 1)
    packet = [header]

    for label in query.name.split(b'.'):
        if len(label) > 63:
            raise ValueError('DNS label too long')

        if len(label) == 0:
            continue

        packet.append(struct.pack(b'>B', len(label)))
        packet.append(label)

    q_class = query.q_class
    if unicast:
        q_class |= 0x8000
    packet.append(struct.pack(b'>BHH', 0, query.q_type, q_class))

    # Add an OPT record to indicate EDNS support
    packet.append(struct.pack(b'>BHHLH', 0, constants.OPT, MAX_PACKET_SIZE,
                              DO, 0))

    return b''.join(packet)



import sys
from setuptools import setup
from setuptools.command.test import test as TestCommand

class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)

with open('README.rst', 'rb') as f:
    long_desc = f.read().decode('utf8')

requires=[]

if sys.version_info < (3,4):
    requires.append('asyncio')

setup(name='asyncdns2',
      version='0.1.0',
      description='Pure Python asynchronous DNS via asyncio',
      long_description=long_desc,
      author='Alastair Houghton',
      author_email='alastair@alastairs-place.net',
      url='https://alastairs-place.net/projects/asyncdns',
      license='MIT License',
      packages=['asyncdns'],
      classifiers=[
          'Development Status :: 4 - Beta',
          'License :: OSI Approved :: MIT License',
          'Topic :: Internet :: Name Service (DNS)',
          'Framework :: AsyncIO'
          ],
      package_data = {
          'asyncdns': ['named.root']
          },
      scripts=['scripts/pydig'],
      install_requires=requires,
      tests_require=['pytest'],
      cmdclass={
          'test': PyTest
          },
      provides=['asyncdns2'])
